@@ -51,6 +51,7 @@
 #include <linux/types.h>
 
 #include <xf86_OSproc.h>
+#include <xserver-properties.h>
 
 #include <unistd.h>
 
@@ -63,16 +64,6 @@
 
 #include <windowstr.h>
 
-#ifdef HAVE_PROPERTIES
-#include <xserver-properties.h>
-/* 1.6 has properties, but no labels */
-#ifdef AXIS_LABEL_PROP
-#define HAVE_LABELS
-#else
-#undef HAVE_LABELS
-#endif
-
-#endif
 
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 23
 #define HAVE_THREADED_INPUT 1
@@ -100,6 +91,7 @@
 #include <qubes-gui-protocol.h>
 #include "xdriver-shm-cmd.h"
 #include "qubes.h"
+#include "labels.h"
 
 #include "../../xf86-qubes-common/include/xf86-qubes-common.h"
 
@@ -125,7 +117,7 @@ static void QubesBlockHandler(void *arg, void *timeout);
 static void QubesWakeupHandler(void *arg, int result);
 #endif
 
-
+#define ArrayLength(x) (sizeof(x)/sizeof((x)[0]))
 
 _X_EXPORT InputDriverRec QUBES = {
     1,
@@ -271,10 +263,34 @@ static void QubesUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 
 static int _qubes_init_kbd(DeviceIntPtr device)
 {
-    InitKeyboardDeviceStruct(device, NULL, NULL, NULL);
+    if (!InitKeyboardDeviceStruct(device, NULL, NULL, NULL))
+        return BadAlloc;
     return Success;
 }
 
+static void QubesInitButtonLabels(QubesDevicePtr pQubes, int natoms,
+                                  Atom * atoms)
+{
+    Atom atom;
+    int btn;
+    const char **labels;
+    int labels_len = 0;
+
+    labels = btn_labels;
+    labels_len = ArrayLength(btn_labels);
+
+    memset(atoms, 0, natoms * sizeof(Atom));
+
+    /* Now fill the ones we know */
+    for (btn = 0; btn < labels_len && btn < natoms; btn++) {
+        atom = XIGetKnownProperty(labels[btn]);
+        if (!atom)    /* Should not happen */
+            continue;
+
+        atoms[btn] = atom;
+    }
+}
+
 static int _qubes_init_buttons(DeviceIntPtr device)
 {
     InputInfoPtr pInfo = device->public.devicePrivate;
@@ -293,6 +309,7 @@ static int _qubes_init_buttons(DeviceIntPtr device)
 
     pQubes->labels = calloc(num_buttons, sizeof(Atom));
 
+    QubesInitButtonLabels(pQubes, num_buttons, pQubes->labels);
     if (!InitButtonClassDeviceStruct(device, num_buttons,
                                      pQubes->labels, map)) {
         xf86Msg(X_ERROR, "%s: Failed to register buttons.\n",
@@ -307,31 +324,24 @@ static int _qubes_init_buttons(DeviceIntPtr device)
 static void QubesInitAxesLabels(QubesDevicePtr pQubes, int natoms,
                                 Atom * atoms)
 {
-#ifdef HAVE_LABELS
     Atom atom;
     int axis;
-    char **labels;
+    const char **labels;
     int labels_len = 0;
-    char *misc_label;
 
     labels = rel_labels;
     labels_len = ArrayLength(rel_labels);
-    misc_label = AXIS_LABEL_PROP_REL_MISC;
 
     memset(atoms, 0, natoms * sizeof(Atom));
 
     /* Now fill the ones we know */
-    for (axis = 0; axis < labels_len; axis++) {
-        if (pQubes->axis_map[axis] == -1)
-            continue;
-
+    for (axis = 0; axis < labels_len && axis < natoms; axis++) {
         atom = XIGetKnownProperty(labels[axis]);
         if (!atom)    /* Should not happen */
             continue;
 
-        atoms[pQubes->axis_map[axis]] = atom;
+        atoms[axis] = atom;
     }
-#endif
 }
 
 
@@ -381,7 +391,7 @@ int connect_unix_socket(QubesDevicePtr pQubes)
     struct sockaddr_un remote;
 
     if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
-        perror("socket");
+        xf86Msg(X_ERROR, "socket(%s): %s\n", pQubes->device, strerror(errno));
         return -1;
     }
 
@@ -390,7 +400,7 @@ int connect_unix_socket(QubesDevicePtr pQubes)
     strncpy(remote.sun_path, pQubes->device, sizeof(remote.sun_path));
     len = strlen(remote.sun_path) + sizeof(remote.sun_family);
     if (connect(s, (struct sockaddr *) &remote, len) == -1) {
-        perror("connect");
+        xf86Msg(X_ERROR, "connect(%s): %s\n", pQubes->device, strerror(errno));
         close(s);
         return -1;
     }
@@ -475,14 +485,22 @@ static int QubesControl(DeviceIntPtr device, int what)
     InputInfoPtr pInfo = device->public.devicePrivate;
     QubesDevicePtr pQubes = pInfo->private;
     DeviceIntPtr master_kbd = NULL;
+    int ret;
 
     switch (what) {
     case DEVICE_INIT:
         device->public.on = FALSE;
-        _qubes_init_buttons(device);
-        _qubes_init_axes(device);
-        _qubes_init_kbd(device);
-        InitPtrFeedbackClassDeviceStruct(device, QubesPtrCtrlProc);
+        ret = _qubes_init_buttons(device);
+        if (ret != Success)
+            return ret;
+        ret = _qubes_init_axes(device);
+        if (ret != Success)
+            return ret;
+        ret = _qubes_init_kbd(device);
+        if (ret != Success)
+            return ret;
+        if (!InitPtrFeedbackClassDeviceStruct(device, QubesPtrCtrlProc))
+            return BadAlloc;
         break;
 
         /* Switch device on.  Establish socket, start event delivery.  */
@@ -497,6 +515,12 @@ static int QubesControl(DeviceIntPtr device, int what)
                         "%s: cannot open device; sleeping...\n",
                         pInfo->name);
                 sleep(1);
+                if (xf86ServerIsExiting()) {
+                    xf86Msg(X_ERROR,
+                            "%s: cannot open device, server exiting, aborting\n",
+                            pInfo->name);
+                    return BadAlloc;
+                }
             }
         } while (pInfo->fd < 0);
 
